DEFAULT_MODEL = "openai:gpt-4o-mini"

__version__ = "0.5.3"
__author__ = "PsiACE, Frost Ming"
__author_email__ = "psiace@apache.org, me@frostming.com"
__copyright__ = "Copyright (c) 2026 Bub Build contributors."
__homepage__ = "https://github.com/bubbuild/republic"
__docs__ = "Minimal, tape-first LLM client with structured outputs."

__all__ = [
    "DEFAULT_MODEL",
    "__author__",
    "__author_email__",
    "__copyright__",
    "__docs__",
    "__homepage__",
    "__version__",
]
